/*
 rotagator.c
 Copyright © 2003 by Padraig Brady <P@draigBrady.com>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE     /* needed to get the defines */
#include <fcntl.h>      /* in glibc 2.2 this has the needed values defined */
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <libgen.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <syslog.h>

static int interactive=0;
static int dir_fd;
static struct sigaction act;
const char* progname;

static void set_notifier(void)
{
    fcntl(dir_fd, F_SETSIG, SIGRTMIN);
    /* DN_ACCESS => (partial) file read in dir
       Note DN_MULTISHOT is not set since multiple
       (queued) signals would be generated for each
       (partial) read, and changing the sig every
       time would be inefficient. */
    fcntl(dir_fd, F_NOTIFY, DN_ACCESS);
}

static void handler(int sig, siginfo_t *si, void *data)
{
}

static void set_handler(void)
{
    act.sa_sigaction = handler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_SIGINFO;
    sigaction(SIGRTMIN, &act, NULL);
}

static void update_sig(char* tmpfile, const char* sig)
{
    char python_cmd[1024];
    snprintf(python_cmd,sizeof(python_cmd),"%s.py %s %s", progname, sig, tmpfile);
    system(python_cmd);
}

static void error(const char *fmt, ...)  __attribute__ ((noreturn,format (printf, 1, 2)));
static void error(const char *fmt, ...)
{
    int n,size = 255;
    char *buff;

    buff = (char *)malloc(size);
    if(!buff) {
        exit(EXIT_FAILURE);
    }

    va_list ap;

    va_start(ap,fmt);
    n = vsnprintf(buff,size,fmt,ap);
    va_end(ap);

    if(size <= n) {
        size = n+1; //As of glibc 2.1 (C99 standard) vsnprintf sets n to required size (not including NULL)
        buff = (char *)realloc(buff,size);
        if(!buff) {
            exit(EXIT_FAILURE);
        }
        va_start(ap,fmt);
        vsnprintf(buff,size,fmt,ap);
        va_end(ap);
    }

    if(!interactive) {
        syslog(LOG_ERR, "%s", buff); //syslog converts \n -> space
    } else {
        fprintf(stderr, "%s\n", buff);
    }

    free(buff);
    exit(EXIT_FAILURE);
}

static void usage(void)  __attribute__ ((noreturn));
static void usage(void)
{
    interactive=1;
    error("Usage: %s [OPTION] /path/to/signature.file\n\t-i\tinteractive, i.e. run in foreground", progname);
}

int main(int argc, char** argv)
{
    char* dir;
    char* tmpfile; const char tmplate[]="/XXXXXX";
    int c;
    progname=basename(argv[0]);

    while ((c=getopt(argc, argv, "i")) != -1) {
        switch(c) {
            case 'i':
                interactive=1;
                break;
            default:
                usage();
                break;
        }
    }

    if (optind != argc-1) {
        usage();
    } else {
        dir=strdup(argv[argc-1]);
        if (!dir) {
            error("strdup(argv[argc-1])");
        }
        dir=dirname(dir);
    }
    tmpfile=(char*)malloc(strlen(dir)+sizeof(tmplate));
    if (!tmpfile) {
        error("malloc(tmpfile)");
    }
    set_handler();
    dir_fd = open(dir, O_RDONLY);
    if (dir_fd<0) {
        error("open(dir(%s)) [%d]",dir,errno);
    }

    if (!interactive) {
        if (daemon(0,0)!=0) {
            interactive=1;
            error("Error [%d] running daemon()", errno);
        }
    }

    while (1) {
        strcpy(tmpfile,dir);
        strcat(tmpfile,tmplate);
        update_sig(tmpfile, argv[argc-1]);
        /* do above at least once so file exists, and isn't stale */
        set_notifier();
        pause();
    }

    free(tmpfile);
    free(dir);
}
